'''
Test cases for the search_studies
'''

# from ..tools.search_studies import search_studies
from aiagents4pharma.talk2cells.agents.scp_agent import get_app
from langchain_core.messages import HumanMessage

def test_agent_scp():
    '''
    Test the agent_scp.
    '''
    unique_id = 12345
    app = get_app(unique_id)
    config = {"configurable": {"thread_id": unique_id}}
    prompt = "Search for studies on Crohns Disease."
    response = app.invoke(
                        {"messages": [HumanMessage(content=prompt)]},
                        config=config
                    )
    assistant_msg = response["messages"][-1].content
    # Check if the assistant message is a string
    assert isinstance(assistant_msg, str)



#!/usr/bin/env python3

"""
Test cases for utils/embeddings/nim_molmim.py
"""

import unittest
from unittest.mock import patch, MagicMock
from ..utils.embeddings.nim_molmim import EmbeddingWithMOLMIM

class TestEmbeddingWithMOLMIM(unittest.TestCase):
    """
    Test cases for EmbeddingWithMOLMIM class.
    """
    def setUp(self):
        self.base_url = "https://fake-nim-api.com/embeddings"
        self.embeddings_model = EmbeddingWithMOLMIM(self.base_url)
        self.test_texts = ["CCO", "CCC", "C=O"]
        self.test_query = "CCO"
        self.mock_response = {
            "embeddings": [[0.1, 0.2, 0.3], [0.4, 0.5, 0.6], [0.7, 0.8, 0.9]]
        }

    @patch("requests.post")
    def test_embed_documents(self, mock_post):
        '''
        Test the embed_documents method.
        '''
        # Mock the response from requests.post
        mock_post.return_value = MagicMock()
        mock_post.return_value.json.return_value = self.mock_response
        embeddings = self.embeddings_model.embed_documents(self.test_texts)
        # Assertions
        self.assertEqual(embeddings, self.mock_response["embeddings"])
        mock_post.assert_called_once_with(
            self.base_url,
            headers={
                'accept': 'application/json',
                'Content-Type': 'application/json'
            },
            data='{"sequences": ["CCO", "CCC", "C=O"]}',
            timeout=60
        )

    @patch("requests.post")
    def test_embed_query(self, mock_post):
        '''
        Test the embed_query method.
        '''
        # Mock the response from requests.post
        mock_post.return_value = MagicMock()
        mock_post.return_value.json.return_value = {"embeddings": [[0.1, 0.2, 0.3]]}
        embedding = self.embeddings_model.embed_query(self.test_query)
        # Assertions
        self.assertEqual(embedding, [[0.1, 0.2, 0.3]])
        mock_post.assert_called_once_with(
            self.base_url,
            headers={
                'accept': 'application/json',
                'Content-Type': 'application/json'
            },
            data='{"sequences": ["CCO"]}',
            timeout=60
        )



'''
Test cases for Talk2Biomodels query_article tool.
'''

from pydantic import BaseModel, Field
from langchain_core.messages import HumanMessage, ToolMessage
from langchain_openai import ChatOpenAI
from langchain_nvidia_ai_endpoints import NVIDIAEmbeddings
from ..agents.t2b_agent import get_app

LLM_MODEL = ChatOpenAI(model='gpt-4o-mini', temperature=0)

class Article(BaseModel):
    '''
    Article schema.
    '''
    title: str = Field(description="Title of the article.")

def test_query_article_with_an_article():
    '''
    Test the query_article tool by providing an article.
    '''
    unique_id = 12345
    app = get_app(unique_id, llm_model=LLM_MODEL)
    config = {"configurable": {"thread_id": unique_id}}
    # Update state by providing the pdf file name
    # and the text embedding model
    app.update_state(config,
      {"pdf_file_name": "aiagents4pharma/talk2biomodels/tests/article_on_model_537.pdf",
       "text_embedding_model": NVIDIAEmbeddings(model='nvidia/llama-3.2-nv-embedqa-1b-v2')})
    prompt = "What is the title of the article?"
    # Test the tool query_article
    response = app.invoke(
                        {"messages": [HumanMessage(content=prompt)]},
                        config=config
                    )
    # Get the response from the tool
    assistant_msg = response["messages"][-1].content
    # Prepare a LLM that can be used as a judge
    llm = LLM_MODEL
    # Make it return a structured output
    structured_llm = llm.with_structured_output(Article)
    # Prepare a prompt for the judge
    prompt = "Given the text below, what is the title of the article?"
    prompt += f"\n\n{assistant_msg}"
    # Get the structured output
    article = structured_llm.invoke(prompt)
    # Check if the article title is correct
    expected_title = "A Multiscale Model of IL-6–Mediated "
    expected_title += "Immune Regulation in Crohn’s Disease"
    # Check if the article title is correct
    assert article.title == expected_title

def test_query_article_without_an_article():
    '''
    Test the query_article tool without providing an article.
    The status of the tool should be error.
    '''
    unique_id = 12345
    app = get_app(unique_id, llm_model=LLM_MODEL)
    config = {"configurable": {"thread_id": unique_id}}
    prompt = "What is the title of the uploaded article?"
    # Update state by providing the text embedding model
    app.update_state(config,
      {"text_embedding_model": NVIDIAEmbeddings(model='nvidia/llama-3.2-nv-embedqa-1b-v2')})
    # Test the tool query_article
    app.invoke(
            {"messages": [HumanMessage(content=prompt)]},
            config=config
        )
    current_state = app.get_state(config)
    # Get the messages from the current state
    # and reverse the order
    reversed_messages = current_state.values["messages"][::-1]
    # Loop through the reversed messages
    # until a ToolMessage is found.
    tool_status_is_error = False
    for msg in reversed_messages:
        if isinstance(msg, ToolMessage):
            # Skip until it finds a ToolMessage
            if msg.name == "query_article" and msg.status == "error":
                tool_status_is_error = True
                break
    assert tool_status_is_error



#!/usr/bin/env python3

"""
Tool for plotting a custom y-axis of a simulation plot.
"""

import logging
from typing import Type, Annotated, List, Tuple, Union, Literal
from pydantic import BaseModel, Field
import hydra
import pandas as pd
from langchain_core.tools import BaseTool
from langchain_core.prompts import ChatPromptTemplate
from langgraph.prebuilt import InjectedState
from .load_biomodel import ModelData, load_biomodel
from .utils import get_model_units

# Initialize logger
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

def extract_relevant_species(question, species_names, state):
    """
    Extract the relevant species from the user question.

    Args:
        question (str): The user question.
        species_names (list): The species names available in the simulation results.
        state (dict): The state of the graph.

    Returns:
        CustomHeader: The relevant species
    """
    # In the following code, we extract the species
    # from the user question. We use Literal to restrict
    # the species names to the ones available in the
    # simulation results.
    class CustomHeader(BaseModel):
        """
        A list of species based on user question.

        This is a Pydantic model that restricts the species
        names to the ones available in the simulation results.
        
        If no species is relevant, set the attribute
        `relevant_species` to None.
        """
        relevant_species: Union[None, List[Literal[*species_names]]] = Field(
                description="This is a list of species based on the user question."
                "It is restricted to the species available in the simulation results."
                "If no species is relevant, set this attribute to None."
                "If the user asks for very specific species (for example, using the"
                "keyword `only` in the question), set this attribute to correspond "
                "to the species available in the simulation results, otherwise set it to None."
                )
    # Load hydra configuration
    with hydra.initialize(version_base=None, config_path="../configs"):
        cfg = hydra.compose(config_name='config',
                            overrides=['tools/custom_plotter=default'])
        cfg = cfg.tools.custom_plotter
    # Get the system prompt
    system_prompt = cfg.system_prompt_custom_header
    # Create an instance of the LLM model
    logging.log(logging.INFO, "LLM model: %s", state['llm_model'])
    llm = state['llm_model']
    llm_with_structured_output = llm.with_structured_output(CustomHeader)
    prompt = ChatPromptTemplate.from_messages([("system", system_prompt),
                                               ("human", "{input}")])
    few_shot_structured_llm = prompt | llm_with_structured_output
    return few_shot_structured_llm.invoke(question)

class CustomPlotterInput(BaseModel):
    """
    Input schema for the custom plotter tool.
    """
    question: str = Field(description="Description of the plot")
    sys_bio_model: ModelData = Field(description="model data",
                                     default=None)
    simulation_name: str = Field(description="Name assigned to the simulation")
    state: Annotated[dict, InjectedState]

# Note: It's important that every field has type hints.
# BaseTool is a Pydantic class and not having type hints
# can lead to unexpected behavior.
# Note: It's important that every field has type hints.
# BaseTool is a Pydantic class and not having type hints
# can lead to unexpected behavior.
class CustomPlotterTool(BaseTool):
    """
    Tool for custom plotting the y-axis of a plot.
    """
    name: str = "custom_plotter"
    description: str = '''A visualization tool designed to extract and display a subset
                        of the larger simulation plot generated by the simulate_model tool.
                        It allows users to specify particular species for the y-axis, 
                        providing a more targeted view of key species without the clutter 
                        of the full plot.'''
    args_schema: Type[BaseModel] = CustomPlotterInput
    response_format: str = "content_and_artifact"

    def _run(self,
             question: str,
             sys_bio_model: ModelData,
             simulation_name: str,
             state: Annotated[dict, InjectedState]
             ) -> Tuple[str, Union[None, List[str]]]:
        """
        Run the tool.

        Args:
            question (str): The question about the custom plot.
            sys_bio_model (ModelData): The model data.
            simulation_name (str): The name assigned to the simulation.
            state (dict): The state of the graph.

        Returns:
            str: The answer to the question
        """
        logger.log(logging.INFO, "Calling custom_plotter tool %s, %s", question, sys_bio_model)
        # Load the model
        sbml_file_path = state['sbml_file_path'][-1] if len(state['sbml_file_path']) > 0 else None
        model_object = load_biomodel(sys_bio_model, sbml_file_path=sbml_file_path)
        dic_simulated_data = {}
        for data in state["dic_simulated_data"]:
            for key in data:
                if key not in dic_simulated_data:
                    dic_simulated_data[key] = []
                dic_simulated_data[key] += [data[key]]
        # Create a pandas dataframe from the dictionary
        df = pd.DataFrame.from_dict(dic_simulated_data)
        # Get the simulated data for the current tool call
        df = pd.DataFrame(
                df[df['name'] == simulation_name]['data'].iloc[0]
                )
        # df = pd.DataFrame.from_dict(state['dic_simulated_data'])
        species_names = df.columns.tolist()
        # Exclude the time column
        species_names.remove('Time')
        logging.log(logging.INFO, "Species names: %s", species_names)
        # Extract the relevant species from the user question
        results = extract_relevant_species(question, species_names, state)
        print (results)
        if results.relevant_species is None:
            raise ValueError("No species found in the simulation results \
                             that matches the user prompt.")
        extracted_species = []
        # Extract the species from the results
        # that are available in the simulation results
        for species in results.relevant_species:
            if species in species_names:
                extracted_species.append(species)
        logging.info("Extracted species: %s", extracted_species)
        # Include the time column
        extracted_species.insert(0, 'Time')
        return f"Custom plot {simulation_name}",{
                            'dic_data': df[extracted_species].to_dict(orient='records')
                            }| get_model_units(model_object)



#!/usr/bin/env python3

"""
BasicoModel class for loading and simulating SBML models
using the basico package.
"""

import logging
from typing import Optional, Dict, Union
from pydantic import Field, model_validator
import pandas as pd
import basico
from .sys_bio_model import SysBioModel

# Initialize logger
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

class BasicoModel(SysBioModel):
    """
    Model that loads and simulates SBML models using the basico package.
    Can load models from an SBML file or download them using a BioModels biomodel_id.
    """
    biomodel_id: Optional[Union[int, str]] = Field(None,
                                description="BioModels model ID to download and load")
    sbml_file_path: Optional[str] = Field(None, description="Path to an SBML file to load")
    simulation_results: Optional[str] = None
    name: Optional[str] = Field("", description="Name of the model")
    description: Optional[str] = Field("", description="Description of the model")

    # Additional attribute not included in the schema
    copasi_model: Optional[object] = None  # Holds the loaded Copasi model

    @model_validator(mode="after")
    def check_biomodel_id_or_sbml_file_path(self):
        """
        Validate that either biomodel_id or sbml_file_path is provided.
        """
        if not self.biomodel_id and not self.sbml_file_path:
            logger.error("Either biomodel_id or sbml_file_path must be provided.")
            raise ValueError("Either biomodel_id or sbml_file_path must be provided.")
        if self.biomodel_id:
            self.copasi_model = basico.load_biomodel(self.biomodel_id)
            self.description = basico.biomodels.get_model_info(self.biomodel_id)["description"]
            self.name = basico.model_info.get_model_name(model=self.copasi_model)
        elif self.sbml_file_path:
            self.copasi_model = basico.load_model(self.sbml_file_path)
            self.description = basico.model_info.get_notes(model=self.copasi_model)
            self.name = basico.model_info.get_model_name(model=self.copasi_model)
        return self

    def update_parameters(self, parameters: Dict[str, Union[float, int]]) -> None:
        """
        Update model parameters with new values.
        """
        # Update parameters in the model
        for param_name, param_value in parameters.items():
            # check if the param_name is not None
            if param_name is None:
                continue
            # Extract all parameters and species from the model
            df_all_params = basico.model_info.get_parameters(model=self.copasi_model)
            df_all_species = basico.model_info.get_species(model=self.copasi_model)
            # if param is a kinetic parameter
            if param_name in df_all_params.index.tolist():
                basico.model_info.set_parameters(name=param_name,
                                            exact=True,
                                            initial_value=param_value,
                                            model=self.copasi_model)
            # if param is a species
            elif param_name in df_all_species.index.tolist():
                basico.model_info.set_species(name=param_name,
                                            exact=True,
                                            initial_concentration=param_value,
                                            model=self.copasi_model)
            else:
                logger.error("Parameter/Species %s not found in the model.", param_name)
                raise ValueError(f"Parameter/Species {param_name} not found in the model.")

    def simulate(self, duration: Union[int, float] = 10, interval: int = 10) -> pd.DataFrame:
        """
        Simulate the COPASI model over a specified range of time points.
        
        Args:
            duration: Duration of the simulation in time units.
            interval: Interval between time points in the simulation.
        
        Returns:
            Pandas DataFrame with time-course simulation results.
        """
        # Run the simulation and return results
        df_result = basico.run_time_course(model=self.copasi_model,
                                        intervals=interval,
                                        duration=duration)
        # # Replace curly braces in column headers with square brackets
        # # Because curly braces in the world of LLMS are used for
        # # structured output
        # df_result.columns = df_result.columns.str.replace('{', '[', regex=False).\
        #             str.replace('}', ']', regex=False)
        # Reset the index
        df_result.reset_index(inplace=True)
        # Store the simulation results
        self.simulation_results = df_result
        # Return copy of the simulation results
        return df_result.copy()

    def get_model_metadata(self) -> Dict[str, Union[str, int]]:
        """
        Retrieve metadata specific to the COPASI model.
        
        Returns:
            Dictionary of model metadata.
        """
        return {
            "Model Type": "SBML Model (COPASI)",
            "Parameter Count": len(basico.get_parameters())
        }



"""
Test cases for agents/t2kg_agent.py
"""

import pytest
from langchain_core.messages import HumanMessage
from langchain_openai import ChatOpenAI, OpenAIEmbeddings
from ..agents.t2kg_agent import get_app

# Define the data path
DATA_PATH = "aiagents4pharma/talk2knowledgegraphs/tests/files"


@pytest.fixture(name="input_dict")
def input_dict_fixture():
    """
    Input dictionary fixture.
    """
    input_dict = {
        "llm_model": None,  # TBA for each test case
        "embedding_model": None,  # TBA for each test case
        "uploaded_files": [
            {
                "file_name": "adalimumab.pdf",
                "file_path": f"{DATA_PATH}/adalimumab.pdf",
                "file_type": "drug_data",
                "uploaded_by": "VPEUser",
                "uploaded_timestamp": "2024-11-05 00:00:00",
            },
            {
                "file_name": "DGE_human_Colon_UC-vs-Colon_Control.pdf",
                "file_path": f"{DATA_PATH}/DGE_human_Colon_UC-vs-Colon_Control.pdf",
                "file_type": "endotype",
                "uploaded_by": "VPEUser",
                "uploaded_timestamp": "2024-11-05 00:00:00",
            },
        ],
        "topk_nodes": 3,
        "topk_edges": 3,
        "dic_source_graph": [
            {
                "name": "PrimeKG",
                "kg_pyg_path": f"{DATA_PATH}/primekg_ibd_pyg_graph.pkl",
                "kg_text_path": f"{DATA_PATH}/primekg_ibd_text_graph.pkl",
            }
        ],
        "dic_extracted_graph": []
    }

    return input_dict


def test_t2kg_agent_openai(input_dict):
    """
    Test the T2KG agent using OpenAI model.

    Args:
        input_dict: Input dictionary
    """
    # Prepare LLM and embedding model
    input_dict["llm_model"] = ChatOpenAI(model="gpt-4o-mini", temperature=0.0)
    input_dict["embedding_model"] = OpenAIEmbeddings(model="text-embedding-3-small")

    # Setup the app
    unique_id = 12345
    app = get_app(unique_id, llm_model=input_dict["llm_model"])
    config = {"configurable": {"thread_id": unique_id}}
    # Update state
    app.update_state(
        config,
        input_dict,
    )
    prompt = """
    Adalimumab is a fully human monoclonal antibody (IgG1) 
    that specifically binds to tumor necrosis factor-alpha (TNF-α), a pro-inflammatory cytokine.

    I would like to get evidence from the knowledge graph about the mechanism of actions related to
    Adalimumab in treating inflammatory bowel disease 
    (IBD). Please follow these steps:
    - Extract a subgraph from the PrimeKG that contains information about Adalimumab.
    - Summarize the extracted subgraph.
    - Reason about the mechanism of action of Adalimumab in treating IBD.

    Please set the extraction name for the extraction process as `subkg_12345`.
    """

    # Test the tool get_modelinfo
    response = app.invoke({"messages": [HumanMessage(content=prompt)]}, config=config)

    # Check assistant message
    assistant_msg = response["messages"][-1].content
    assert isinstance(assistant_msg, str)

    # Check extracted subgraph dictionary
    current_state = app.get_state(config)
    dic_extracted_graph = current_state.values["dic_extracted_graph"][0]
    assert isinstance(dic_extracted_graph, dict)
    assert dic_extracted_graph["name"] == "subkg_12345"
    assert dic_extracted_graph["graph_source"] == "PrimeKG"
    assert dic_extracted_graph["topk_nodes"] == 3
    assert dic_extracted_graph["topk_edges"] == 3
    assert isinstance(dic_extracted_graph["graph_dict"], dict)
    assert len(dic_extracted_graph["graph_dict"]["nodes"]) > 0
    assert len(dic_extracted_graph["graph_dict"]["edges"]) > 0
    assert isinstance(dic_extracted_graph["graph_text"], str)
    # Check summarized subgraph
    assert isinstance(dic_extracted_graph["graph_summary"], str)
    # Check reasoning results
    assert "Adalimumab" in assistant_msg
    assert "TNF" in assistant_msg



'''
This is the agent file for the Talk2KnowledgeGraphs agent.
'''

import logging
from typing import Annotated
import hydra
from langchain_core.language_models.chat_models import BaseChatModel
from langgraph.checkpoint.memory import MemorySaver
from langgraph.graph import START, StateGraph
from langgraph.prebuilt import create_react_agent, ToolNode, InjectedState
from ..tools.subgraph_extraction import SubgraphExtractionTool
from ..tools.subgraph_summarization import SubgraphSummarizationTool
from ..tools.graphrag_reasoning import GraphRAGReasoningTool
from ..states.state_talk2knowledgegraphs import Talk2KnowledgeGraphs

# Initialize logger
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

def get_app(uniq_id, llm_model: BaseChatModel):
    '''
    This function returns the langraph app.
    '''
    def agent_t2kg_node(state: Annotated[dict, InjectedState]):
        '''
        This function calls the model.
        '''
        logger.log(logging.INFO, "Calling t2kg_agent node with thread_id %s", uniq_id)
        response = model.invoke(state, {"configurable": {"thread_id": uniq_id}})

        return response

    # Load hydra configuration
    logger.log(logging.INFO, "Load Hydra configuration for Talk2KnowledgeGraphs agent.")
    with hydra.initialize(version_base=None, config_path="../configs"):
        cfg = hydra.compose(config_name='config',
                            overrides=['agents/t2kg_agent=default'])
        cfg = cfg.agents.t2kg_agent

    # Define the tools
    subgraph_extraction = SubgraphExtractionTool()
    subgraph_summarization = SubgraphSummarizationTool()
    graphrag_reasoning = GraphRAGReasoningTool()
    tools = ToolNode([
                    subgraph_extraction,
                    subgraph_summarization,
                    graphrag_reasoning,
                    ])

    # Create the agent
    model = create_react_agent(
                llm_model,
                tools=tools,
                state_schema=Talk2KnowledgeGraphs,
                prompt=cfg.state_modifier,
                version='v2',
                checkpointer=MemorySaver()
            )

    # Define a new graph
    workflow = StateGraph(Talk2KnowledgeGraphs)

    # Define the two nodes we will cycle between
    workflow.add_node("agent_t2kg", agent_t2kg_node)

    # Set the entrypoint as the first node
    # This means that this node is the first one called
    workflow.add_edge(START, "agent_t2kg")

    # Initialize memory to persist state between graph runs
    checkpointer = MemorySaver()

    # Finally, we compile it!
    # This compiles it into a LangChain Runnable,
    # meaning you can use it as you would any other runnable.
    # Note that we're (optionally) passing the memory
    # when compiling the graph
    app = workflow.compile(checkpointer=checkpointer,
                           name="T2KG_Agent")
    logger.log(logging.INFO,
               "Compiled the graph with thread_id %s and llm_model %s",
               uniq_id,
               llm_model)

    return app



#!/usr/bin/env python3


"""
Tool for displaying search or recommendation results.

This module defines a tool that retrieves and displays a table of research papers
found during searches or recommendations. If no papers are found, an exception is raised
to signal the need for a new search.
"""


import logging

from typing import Annotated
from langchain_core.messages import ToolMessage
from langchain_core.tools import tool
from langchain_core.tools.base import InjectedToolCallId
from langgraph.prebuilt import InjectedState
from langgraph.types import Command


# Configure logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class NoPapersFoundError(Exception):
    """
    Exception raised when no research papers are found in the agent's state.

    This exception helps the language model determine whether a new search
    or recommendation should be initiated.

    Example:
        >>> if not papers:
        >>>     raise NoPapersFoundError("No papers found. A search is needed.")
    """


@tool("display_results", parse_docstring=True)
def display_results(
    tool_call_id: Annotated[str, InjectedToolCallId],
    state: Annotated[dict, InjectedState],
) -> Command:
    """
    Displays retrieved research papers after a search or recommendation.

    This function retrieves the last displayed research papers from the state and
    returns them as an artifact for further processing. If no papers are found,
    it raises a `NoPapersFoundError` to indicate that a new search is needed.

    Args:
        tool_call_id (Annotated[str, InjectedToolCallId]): The tool call ID for tracking.
        state (dict): The agent's state containing retrieved papers.

    Returns:
        Command: A command containing a message with the number of displayed papers
                 and an attached artifact for further reference.

    Raises:
        NoPapersFoundError: If no research papers are found in the agent's state.

    Example:
        >>> state = {"last_displayed_papers": {"paper1": "Title 1", "paper2": "Title 2"}}
        >>> result = display_results(tool_call_id="123", state=state)
        >>> print(result.update["messages"][0].content)
        "2 papers found. Papers are attached as an artifact."
    """
    logger.info("Displaying papers")
    context_key = state.get("last_displayed_papers")
    artifact = state.get(context_key)
    if not artifact:
        logger.info("No papers found in state, raising NoPapersFoundError")
        raise NoPapersFoundError(
            "No papers found. A search/rec needs to be performed first."
        )
    content = f"{len(artifact)} papers found. Papers are attached as an artifact."
    return Command(
        update={
            "messages": [
                ToolMessage(
                    content=content,
                    tool_call_id=tool_call_id,
                    artifact=artifact,
                )
            ],
        }
    )



"""
Embedding class using HuggingFace model based on LangChain Embeddings class.
"""

from typing import List
import torch
from transformers import AutoModel, AutoTokenizer, AutoConfig
from .embeddings import Embeddings

class EmbeddingWithHuggingFace(Embeddings):
    """
    Embedding class using HuggingFace model based on LangChain Embeddings class.
    """

    def __init__(
        self,
        model_name: str,
        model_cache_dir: str = None,
        truncation: bool = True,
        device: str = "cpu",
    ):
        """
        Initialize the EmbeddingWithHuggingFace class.

        Args:
            model_name: The name of the HuggingFace model to be used.
            model_cache_dir: The directory to cache the HuggingFace model.
            truncation: The truncation flag for the HuggingFace tokenizer.
            return_tensors: The return_tensors flag for the HuggingFace tokenizer.
            device: The device to run the model on.
        """

        # Set parameters
        self.model_name = model_name
        self.model_cache_dir = model_cache_dir
        self.truncation = truncation
        self.device = device

        # Try to load the model from HuggingFace Hub
        try:
            AutoConfig.from_pretrained(self.model_name)
        except EnvironmentError as e:
            raise ValueError(
                f"Model {self.model_name} is not available on HuggingFace Hub."
            ) from e

        # Load HuggingFace tokenizer and model
        self.tokenizer = AutoTokenizer.from_pretrained(
            self.model_name, cache_dir=self.model_cache_dir
        )
        self.model = AutoModel.from_pretrained(
            self.model_name, cache_dir=self.model_cache_dir
        )

    def meanpooling(self, output, mask) -> torch.Tensor:
        """
        Mean Pooling - Take attention mask into account for correct averaging.
        According to the following documentation:
        https://huggingface.co/NeuML/pubmedbert-base-embeddings

        Args:
            output: The output of the model.
            mask: The mask of the model.
        """
        embeddings = output[0] # First element of model_output contains all token embeddings
        mask = mask.unsqueeze(-1).expand(embeddings.size()).float()
        return torch.sum(embeddings * mask, 1) / torch.clamp(mask.sum(1), min=1e-9)

    def embed_documents(self, texts: List[str]) -> List[float]:
        """
        Generate embedding for a list of input texts using HuggingFace model.

        Args:
            texts: The list of texts to be embedded.

        Returns:
            The list of embeddings for the given texts.
        """

        # Generate the embedding
        with torch.no_grad():
            inputs = self.tokenizer(
                texts,
                padding=True,
                truncation=self.truncation,
                return_tensors="pt",
            ).to(self.device)
            outputs = self.model.to(self.device)(**inputs)
            embeddings = self.meanpooling(outputs, inputs['attention_mask']).cpu()

        return embeddings

    def embed_query(self, text: str) -> List[float]:
        """
        Generate embeddings for an input text using HuggingFace model.

        Args:
            text: A query to be embedded.
        Returns:
            The embeddings for the given query.
        """

        # Generate the embedding
        with torch.no_grad():
            inputs = self.tokenizer(
                text,
                padding=True,
                truncation=self.truncation,
                return_tensors="pt",
            ).to(self.device)
            outputs = self.model.to(self.device)(**inputs)
            embeddings = self.meanpooling(outputs, inputs['attention_mask']).cpu()[0]

        return embeddings



"""
Embedding class using Ollama model based on LangChain Embeddings class.
"""

import time
from typing import List
import subprocess
import ollama
from langchain_ollama import OllamaEmbeddings
from .embeddings import Embeddings

class EmbeddingWithOllama(Embeddings):
    """
    Embedding class using Ollama model based on LangChain Embeddings class.
    """
    def __init__(self, model_name: str):
        """
        Initialize the EmbeddingWithOllama class.

        Args:
            model_name: The name of the Ollama model to be used.
        """
        # Setup the Ollama server
        self.__setup(model_name)

        # Set parameters
        self.model_name = model_name

        # Prepare model
        self.model = OllamaEmbeddings(model=self.model_name)

    def __setup(self, model_name: str) -> None:
        """
        Check if the Ollama model is available and run the Ollama server if needed.

        Args:
            model_name: The name of the Ollama model to be used.
        """
        try:
            models_list = ollama.list()["models"]
            if model_name not in [m['model'].replace(":latest", "") for m in models_list]:
                ollama.pull(model_name)
                time.sleep(30)
                raise ValueError(f"Pulled {model_name} model")
        except Exception as e:
            with subprocess.Popen(
                "ollama serve", shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
            ):
                time.sleep(10)
            raise ValueError(f"Error: {e} and restarted Ollama server.") from e

    def embed_documents(self, texts: List[str]) -> List[float]:
        """
        Generate embedding for a list of input texts using Ollama model.

        Args:
            texts: The list of texts to be embedded.

        Returns:
            The list of embeddings for the given texts.
        """

        # Generate the embedding
        embeddings = self.model.embed_documents(texts)

        return embeddings

    def embed_query(self, text: str) -> List[float]:
        """
        Generate embeddings for an input text using Ollama model.

        Args:
            text: A query to be embedded.
        Returns:
            The embeddings for the given query.
        """

        # Generate the embedding
        embeddings = self.model.embed_query(text)

        return embeddings



"""
Tests for state management functionality.
"""

from ..state.state_talk2scholars import replace_dict


def test_state_replace_dict():
    """Verifies state dictionary replacement works correctly"""
    existing = {"key1": "value1", "key2": "value2"}
    new = {"key3": "value3"}
    result = replace_dict(existing, new)
    assert result == new
    assert isinstance(result, dict)



'''
Test cases for Talk2Biomodels.
'''

from langchain_core.messages import HumanMessage
from langchain_openai import ChatOpenAI
from ..agents.t2b_agent import get_app

LLM_MODEL = ChatOpenAI(model='gpt-4o-mini', temperature=0)

def test_simulate_model_tool():
    '''
    Test the simulate_model tool when simulating
    multiple models.
    '''
    unique_id = 123
    app = get_app(unique_id, llm_model=LLM_MODEL)
    config = {"configurable": {"thread_id": unique_id}}
    # Upload a model to the state
    app.update_state(config,
        {"sbml_file_path": ["aiagents4pharma/talk2biomodels/tests/BIOMD0000000449_url.xml"]})
    prompt = "Simulate model 64 and the uploaded model"
    # Invoke the agent
    app.invoke(
        {"messages": [HumanMessage(content=prompt)]},
        config=config
    )
    current_state = app.get_state(config)
    dic_simulated_data = current_state.values["dic_simulated_data"]
    # Check if the dic_simulated_data is a list
    assert isinstance(dic_simulated_data, list)
    # Check if the length of the dic_simulated_data is 2
    assert len(dic_simulated_data) == 2
    # Check if the source of the first model is 64
    assert dic_simulated_data[0]['source'] == 64
    # Check if the source of the second model is upload
    assert dic_simulated_data[1]['source'] == "upload"
    # Check if the data of the first model contains
    assert '1,3-bisphosphoglycerate' in dic_simulated_data[0]['data']
    # Check if the data of the second model contains
    assert 'mTORC2' in dic_simulated_data[1]['data']



"""
This is the state file for the Talk2AIAgents4Pharma agent.
"""

from ...talk2biomodels.states.state_talk2biomodels import Talk2Biomodels
from ...talk2knowledgegraphs.states.state_talk2knowledgegraphs import Talk2KnowledgeGraphs

class Talk2AIAgents4Pharma(Talk2Biomodels,
                           Talk2KnowledgeGraphs):
    """
    The state for the Talk2AIAgents4Pharma agent.

    This class inherits from the classes:
    1. Talk2Biomodels
    2. Talk2KnowledgeGraphs
    """



'''
Test cases for Talk2Biomodels search models tool.
'''

from langchain_core.messages import HumanMessage, ToolMessage
from langchain_openai import ChatOpenAI
from ..agents.t2b_agent import get_app

LLM_MODEL = ChatOpenAI(model='gpt-4o-mini', temperature=0)

def test_search_models_tool():
    '''
    Test the search_models tool.
    '''
    unique_id = 12345
    app = get_app(unique_id, llm_model=LLM_MODEL)
    config = {"configurable": {"thread_id": unique_id}}
    prompt = "Search for models on Crohn's disease."
    app.update_state(
            config,
            {"llm_model": LLM_MODEL}
        )
    # Test the tool get_modelinfo
    response = app.invoke(
                        {"messages": [HumanMessage(content=prompt)]},
                        config=config
                    )
    # Extract the assistant artifact which contains
    # all the search results
    found_model_537 = False
    for msg in response["messages"]:
        if isinstance(msg, ToolMessage) and msg.name == "search_models":
            msg_artifact = msg.artifact
            for model in msg_artifact["dic_data"]:
                if model["id"] == "BIOMD0000000537":
                    found_model_537 = True
            break
    # Check if the model BIOMD0000000537 is found
    assert found_model_537



#!/usr/bin/env python3

'''
A tool to fetch studies from the Single Cell Portal.
'''

import logging
from typing import Annotated
import requests
from langchain_core.tools import tool
from langchain_core.tools.base import InjectedToolCallId
from langchain_core.messages import ToolMessage
from langgraph.types import Command
import pandas as pd

# Initialize logger
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

@tool('search_studies')
def search_studies(search_term: str,
                   tool_call_id: Annotated[str, InjectedToolCallId],
                   limit: int = 5):
    """
    Fetch studies from single cell portal

    Args:
        search_term (str): The search term to use. Example: "COVID-19", "cancer", etc.
        limit (int): The number of papers to return. Default is 5.

    """
    logger.log(logging.INFO, "Calling the tool search_studies")
    scp_endpoint = 'https://singlecell.broadinstitute.org/single_cell/api/v1/search?type=study'
    # params = {'terms': search_term, 'facets': 'MONDO_0005011'}
    params = {'terms': search_term}
    status_code = 0
    while status_code != 200:
        # Make a GET request to the single cell portal
        search_response = requests.get(scp_endpoint,
                                       params=params,
                                       timeout=10,
                                       verify=False)
        status_code = search_response.status_code
        logger.log(logging.INFO, "Status code %s received from SCP")

    # Select the columns to display in the table
    selected_columns = ["study_source", "name", "study_url", "gene_count", "cell_count"]

    # Extract the data from the response
    # with the selected columns
    df = pd.DataFrame(search_response.json()['studies'])[selected_columns]

    # Convert column 'Study Name' into clickable
    # hyperlinks from the column 'Study URL'
    scp_api_url = 'https://singlecell.broadinstitute.org'
    df['name'] = df.apply(
            lambda x: f"<a href=\"{scp_api_url}/{x['study_url']}\">{x['name']}</a>",
            axis=1)

    # Excldue the column 'Study URL' from the dataframe
    df = df.drop(columns=['study_url'])

    # Add a new column a the beginning of the dataframe with row numbers
    df.insert(0, 'S/N', range(1, 1 + len(df)))

    # Update the state key 'search_table' with the dataframe in markdown format
    return Command(
        update={
            # update the state keys
            "search_table": df.to_markdown(tablefmt="grid"),
            # update the message history
            "messages": [
                ToolMessage(
                    f"Successfully fetched {limit} studies on {search_term}.",
                    tool_call_id=tool_call_id
                )
            ],
        }
    )



"""
Unit tests for Zotero search tool in zotero_read.py.
"""

from unittest.mock import patch
from langgraph.types import Command
from ..tools.zotero.zotero_read import zotero_search_tool


# Mock data for Zotero API response
MOCK_ZOTERO_RESPONSE = [
    {
        "data": {
            "key": "ABC123",
            "title": "Deep Learning in Medicine",
            "abstractNote": "An overview of deep learning applications in medicine.",
            "date": "2022",
            "url": "https://example.com/paper1",
            "itemType": "journalArticle",
        }
    },
    {
        "data": {
            "key": "XYZ789",
            "title": "Advances in AI",
            "abstractNote": "Recent advancements in AI research.",
            "date": "2023",
            "url": "https://example.com/paper2",
            "itemType": "conferencePaper",
        }
    },
]


class TestZoteroRead:
    """Unit tests for Zotero search tool"""

    @patch("pyzotero.zotero.Zotero.items")
    def test_zotero_success(self, mock_zotero):
        """Verifies successful retrieval of papers from Zotero"""
        mock_zotero.return_value = MOCK_ZOTERO_RESPONSE

        result = zotero_search_tool.invoke(
            input={
                "query": "deep learning",
                "only_articles": True,
                "limit": 2,
                "tool_call_id": "test123",
            }
        )

        assert isinstance(result, Command)
        assert "zotero_read" in result.update
        assert "messages" in result.update
        papers = result.update["zotero_read"]
        assert len(papers) == 2  # Should return 2 papers
        assert papers["ABC123"]["Title"] == "Deep Learning in Medicine"
        assert papers["XYZ789"]["Title"] == "Advances in AI"

    @patch("pyzotero.zotero.Zotero.items")
    def test_zotero_no_papers_found(self, mock_zotero):
        """Verifies Zotero tool behavior when no papers are found"""
        mock_zotero.return_value = []  # Simulating empty response

        result = zotero_search_tool.invoke(
            input={
                "query": "nonexistent topic",
                "only_articles": True,
                "limit": 2,
                "tool_call_id": "test123",
            }
        )

        assert isinstance(result, Command)
        assert "zotero_read" in result.update
        assert len(result.update["zotero_read"]) == 0  # No papers found
        assert "messages" in result.update
        assert "Number of papers found: 0" in result.update["messages"][0].content

    @patch("pyzotero.zotero.Zotero.items")
    def test_zotero_only_articles_filtering(self, mock_zotero):
        """Ensures only journal articles and conference papers are returned"""
        mock_response = [
            {
                "data": {
                    "key": "DEF456",
                    "title": "A Book on AI",
                    "abstractNote": "Book about AI advancements.",
                    "date": "2021",
                    "url": "https://example.com/book",
                    "itemType": "book",
                }
            },
            MOCK_ZOTERO_RESPONSE[0],  # Valid journal article
        ]
        mock_zotero.return_value = mock_response

        result = zotero_search_tool.invoke(
            input={
                "query": "AI",
                "only_articles": True,
                "limit": 2,
                "tool_call_id": "test123",
            }
        )

        assert isinstance(result, Command)
        assert "zotero_read" in result.update
        papers = result.update["zotero_read"]
        assert len(papers) == 1  # The book should be filtered out
        assert "ABC123" in papers  # Journal article should be included

    @patch("pyzotero.zotero.Zotero.items")
    def test_zotero_invalid_response(self, mock_zotero):
        """Tests handling of malformed API response"""
        mock_zotero.return_value = [
            {"data": None},  # Invalid response format
            {},  # Empty object
            {"data": {"title": "Missing Key", "itemType": "journalArticle"}},  # No key
        ]

        result = zotero_search_tool.invoke(
            input={
                "query": "AI ethics",
                "only_articles": True,
                "limit": 2,
                "tool_call_id": "test123",
            }
        )

        assert isinstance(result, Command)
        assert "zotero_read" in result.update
        assert (
            len(result.update["zotero_read"]) == 0
        )  # Should filter out invalid entries
        assert "messages" in result.update
        assert "Number of papers found: 0" in result.update["messages"][0].content

    @patch("pyzotero.zotero.Zotero.items")
    def test_zotero_handles_non_dict_items(self, mock_zotero):
        """Ensures that Zotero tool correctly skips non-dictionary items (covers line 86)"""

        # Simulate Zotero returning an invalid item (e.g., `None` and a string)
        mock_zotero.return_value = [
            None,
            "invalid_string",
            {
                "data": {
                    "key": "123",
                    "title": "Valid Paper",
                    "itemType": "journalArticle",
                }
            },
        ]

        result = zotero_search_tool.invoke(
            input={
                "query": "AI ethics",
                "only_articles": True,
                "limit": 2,
                "tool_call_id": "test123",
            }
        )

        assert isinstance(result, Command)
        assert "zotero_read" in result.update

        # Expect only valid items to be processed
        assert (
            len(result.update["zotero_read"]) == 1
        ), "Only valid dictionary items should be processed"



"""
State management for the Talk2Scholars agent.

This module defines the state class `Talk2Scholars`, which maintains the conversation
context, retrieved papers, and other relevant metadata. The state ensures consistency
across agent interactions.
"""

import logging
from typing import Annotated, Any, Dict
from langchain_core.language_models import BaseChatModel
from langgraph.prebuilt.chat_agent_executor import AgentState

# Configure logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def replace_dict(existing: Dict[str, Any], new: Dict[str, Any]) -> Dict[str, Any]:
    """
    Replaces the existing dictionary with a new dictionary.

    This function logs the state update and ensures that the old state is replaced
    with the new one.

    Args:
        existing (Dict[str, Any]): The current dictionary state.
        new (Dict[str, Any]): The new dictionary state to replace the existing one.

    Returns:
        Dict[str, Any]: The updated dictionary state.

    Example:
        >>> old_state = {"papers": {"id1": "Paper 1"}}
        >>> new_state = {"papers": {"id2": "Paper 2"}}
        >>> updated_state = replace_dict(old_state, new_state)
        >>> print(updated_state)
        {"papers": {"id2": "Paper 2"}}
    """
    logger.info("Updating existing state %s with the state dict: %s", existing, new)
    return new


class Talk2Scholars(AgentState):
    """
    Represents the state of the Talk2Scholars agent.

    This class extends `AgentState` to maintain conversation history, retrieved papers,
    and interactions with the language model.

    Attributes:
        last_displayed_papers (Dict[str, Any]): Stores the most recently displayed papers.
        papers (Dict[str, Any]): Stores the research papers retrieved from the agent's queries.
        multi_papers (Dict[str, Any]): Stores multiple recommended papers from various sources.
        zotero_read (Dict[str, Any]): Stores the papers retrieved from Zotero.
        llm_model (BaseChatModel): The language model instance used for generating responses.
    """

    # Agent state fields
    last_displayed_papers: Annotated[Dict[str, Any], replace_dict]
    papers: Annotated[Dict[str, Any], replace_dict]
    multi_papers: Annotated[Dict[str, Any], replace_dict]
    zotero_read: Annotated[Dict[str, Any], replace_dict]
    llm_model: BaseChatModel



#!/usr/bin/env python3

"""
Tool for parameter scan.
"""

import logging
from typing import Type, Annotated
import basico
from pydantic import BaseModel, Field
from langgraph.types import Command
from langgraph.prebuilt import InjectedState
from langchain_core.tools import BaseTool
from langchain_core.messages import ToolMessage
from langchain_core.tools.base import InjectedToolCallId
from .load_biomodel import ModelData, load_biomodel
from .load_arguments import ArgumentData, add_rec_events

# Initialize logger
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

def run_steady_state(model_object,
                     dic_species_to_be_analyzed_before_experiment):
    """
    Run the steady state analysis.

    Args:
        model_object: The model object.
        dic_species_to_be_analyzed_before_experiment: Dictionary of species data.

    Returns:
        DataFrame: The results of the steady state analysis.
    """
    # Update the fixed model species and parameters
    # These are the initial conditions of the model
    # set by the user
    model_object.update_parameters(dic_species_to_be_analyzed_before_experiment)
    logger.log(logging.INFO, "Running steady state analysis")
    # Run the steady state analysis
    output = basico.task_steadystate.run_steadystate(model=model_object.copasi_model)
    if output == 0:
        logger.error("Steady state analysis failed")
        raise ValueError("A steady state was not found")
    logger.log(logging.INFO, "Steady state analysis successful")
    # Store the steady state results in a DataFrame
    df_steady_state = basico.model_info.get_species(model=model_object.copasi_model).reset_index()
    # print (df_steady_state)
    # Rename the column name to species_name
    df_steady_state.rename(columns={'name': 'species_name'},
                           inplace=True)
    # Rename the column concentration to steady_state_concentration
    df_steady_state.rename(columns={'concentration': 'steady_state_concentration'},
                           inplace=True)
    # Rename the column transition_time to steady_state_transition_time
    df_steady_state.rename(columns={'transition_time': 'steady_state_transition_time'},
                           inplace=True)
    # Drop some columns
    df_steady_state.drop(columns=
                         [
                            'initial_particle_number',
                            'initial_expression',
                            'expression',
                            'particle_number',
                            'type',
                            'particle_number_rate',
                            'key',
                            'sbml_id',
                            'display_name'],
                            inplace=True)
    logger.log(logging.INFO, "Steady state results with shape %s", df_steady_state.shape)
    return df_steady_state

class SteadyStateInput(BaseModel):
    """
    Input schema for the steady state tool.
    """
    sys_bio_model: ModelData = Field(description="model data",
                                     default=None)
    arg_data: ArgumentData = Field(
        description="time, species, and reocurring data"
                " that must be set before the steady state analysis"
                " as well as the experiment name", default=None)
    tool_call_id: Annotated[str, InjectedToolCallId]
    state: Annotated[dict, InjectedState]

# Note: It's important that every field has type hints. BaseTool is a
# Pydantic class and not having type hints can lead to unexpected behavior.
class SteadyStateTool(BaseTool):
    """
    Tool to bring a model to steady state.
    """
    name: str = "steady_state"
    description: str = "A tool to bring a model to steady state."
    args_schema: Type[BaseModel] = SteadyStateInput

    def _run(self,
        tool_call_id: Annotated[str, InjectedToolCallId],
        state: Annotated[dict, InjectedState],
        sys_bio_model: ModelData = None,
        arg_data: ArgumentData = None
    ) -> Command:
        """
        Run the tool.

        Args:
            tool_call_id (str): The tool call ID. This is injected by the system.
            state (dict): The state of the tool.
            sys_bio_model (ModelData): The model data.
            arg_data (ArgumentData): The argument data.

        Returns:
            Command: The updated state of the tool.
        """
        logger.log(logging.INFO, "Calling the steady_state tool %s, %s",
                   sys_bio_model, arg_data)
        # print (f'Calling steady_state tool {sys_bio_model}, {arg_data}, {tool_call_id}')
        sbml_file_path = state['sbml_file_path'][-1] if len(state['sbml_file_path']) > 0 else None
        model_object = load_biomodel(sys_bio_model,
                                  sbml_file_path=sbml_file_path)
        # Prepare the dictionary of species data
        # that will be passed to the simulate method
        # of the BasicoModel class
        dic_species_to_be_analyzed_before_experiment = {}
        if arg_data:
            # Prepare the dictionary of species data
            if arg_data.species_to_be_analyzed_before_experiment is not None:
                dic_species_to_be_analyzed_before_experiment = dict(
                    zip(arg_data.species_to_be_analyzed_before_experiment.species_name,
                        arg_data.species_to_be_analyzed_before_experiment.species_concentration))
            # Add reocurring events (if any) to the model
            if arg_data.reocurring_data is not None:
                add_rec_events(model_object, arg_data.reocurring_data)
        # Run the parameter scan
        df_steady_state = run_steady_state(model_object,
                                           dic_species_to_be_analyzed_before_experiment)
        print (df_steady_state)
        # Prepare the dictionary of scanned data
        # that will be passed to the state of the graph
        dic_steady_state_data = {
            'name': arg_data.experiment_name,
            'source': sys_bio_model.biomodel_id if sys_bio_model.biomodel_id else 'upload',
            'tool_call_id': tool_call_id,
            'data': df_steady_state.to_dict(orient='records')
        }
        # Prepare the dictionary of updated state for the model
        dic_updated_state_for_model = {}
        for key, value in {
            "model_id": [sys_bio_model.biomodel_id],
            "sbml_file_path": [sbml_file_path],
            "dic_steady_state_data": [dic_steady_state_data]
            }.items():
            if value:
                dic_updated_state_for_model[key] = value
        # Return the updated state
        return Command(
                update=dic_updated_state_for_model|{
                # Update the message history
                "messages": [
                ToolMessage(
                        content=f"Steady state analysis of"
                                f" {arg_data.experiment_name}"
                                " was successful.",
                        tool_call_id=tool_call_id,
                        artifact={'dic_data': df_steady_state.to_dict(orient='records')}
                        )
                    ],
                }
            )
